from pathlib import Path
from typing import Iterator

from lxml import etree as ElementTree

from whispers.core.utils import global_exception_handler
from whispers.models.pair import KeyValuePair
from whispers.plugins.common import Common


class Xml:
    def __init__(self):
        self.keypath = []

    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        def _traverse(tree):
            """Traverse XML document"""
            for event, element in tree:
                if event == "end":
                    self.keypath.pop()
                    continue

                self.keypath.append(element.tag)

                # Format: <elem key="value">
                for key, value in element.attrib.items():
                    self.keypath.append(key)
                    yield KeyValuePair(key, value, list(self.keypath))

                    # Format: <elem name="jdbc:mysql://host?k1=v1&amp;k2=v2">
                    yield from Common(self.keypath).pairs(value)

                    self.keypath.pop()

                # Format: <elem key=a value=b>
                if "key" in element.attrib and "value" in element.attrib:
                    yield KeyValuePair(element.attrib["key"], element.attrib["value"], list(self.keypath))

                # Format: <key>value</key>
                if not element.text:
                    continue

                yield KeyValuePair(element.tag, element.text, list(self.keypath))
                yield from Common(self.keypath).pairs(element.text)

                # Format: <elem>key=value</elem>
                if "=" in element.text:
                    item = element.text.split("=")
                    if len(item) == 2:
                        self.keypath.append(item[0])
                        yield KeyValuePair(item[0], item[1], list(self.keypath))
                        self.keypath.pop()

                # Format: <key>name</key><value>string</value>
                found_key = None
                found_value = None
                for item in element:
                    if str(item.tag).lower() == "key":
                        found_key = item.text
                    elif str(item.tag).lower() == "value":
                        found_value = item.text

                if found_key and found_value:
                    self.keypath.append(found_key)
                    yield KeyValuePair(found_key, found_value, list(self.keypath))
                    yield from Common(self.keypath).pairs(found_value)
                    self.keypath.pop()

        try:
            parser = ElementTree.XMLParser(recover=True)
            tree = ElementTree.parse(filepath.as_posix(), parser)
            tree = ElementTree.iterwalk(tree, events=("start", "end"))
            yield from _traverse(tree)

        except Exception:
            global_exception_handler(filepath.as_posix(), tree)



from typing import Iterator


def secrets(arguments: str) -> Iterator:
    """
    Boilerplate for simplifying usage and integration.

    import whispers
    for secret in whispers.secrets("-r apikey tests/fixtures"):
        print(secret)
    """
    import shlex

    from whispers.core.args import parse_args
    from whispers.main import run

    argv = shlex.split(arguments)
    return run(parse_args(argv))



import logging
from pathlib import Path
from typing import Iterator, Optional

from whispers.core.constants import REGEX_PRIVKEY_FILE
from whispers.core.utils import global_exception_handler, is_static, strip_string
from whispers.models.appconfig import AppConfig
from whispers.models.pair import KeyValuePair
from whispers.plugins.config import Config
from whispers.plugins.dockercfg import Dockercfg
from whispers.plugins.dockerfile import Dockerfile
from whispers.plugins.gradle import Gradle
from whispers.plugins.html import Html
from whispers.plugins.htpasswd import Htpasswd
from whispers.plugins.jproperties import Jproperties
from whispers.plugins.json import Json
from whispers.plugins.npmrc import Npmrc
from whispers.plugins.pip import Pip
from whispers.plugins.plaintext import Plaintext
from whispers.plugins.pypirc import Pypirc
from whispers.plugins.python import Python
from whispers.plugins.shell import Shell
from whispers.plugins.xml import Xml
from whispers.plugins.yml import Yml


def make_pairs(config: dict, file: Path) -> Optional[Iterator[KeyValuePair]]:
    """Generates KeyValuePair objects by parsing given file"""
    try:
        if not file.exists():
            return None

        if not file.is_file():
            return None

    except Exception:  # pragma: no cover
        global_exception_handler(file.as_posix(), "make_pairs()")
        return None

    # First, return file name to check if it is a sensitive file
    pair = KeyValuePair("file", file.as_posix())
    if filter_included(config, pair):
        yield tag_file(file, pair)

    # Second, attempt to parse the file with a plugin
    plugin = load_plugin(file)

    logging.debug(f"make_pairs '{plugin}' for '{file}'")

    if not plugin:
        return None

    pairs = plugin().pairs(file)
    static = filter(None, map(filter_static, pairs))
    included = filter(None, map(lambda pair: filter_included(config, pair), static))
    tagged = map(lambda pair: tag_file(file, pair), included)

    try:
        yield from tagged

    except Exception:  # pragma: no cover
        global_exception_handler(file.as_posix(), "make_pairs()")
        return None


def tag_file(file: Path, pair: KeyValuePair) -> KeyValuePair:
    """Add pair file path"""
    pair.file = file.as_posix()
    return pair


def filter_included(config: AppConfig, pair: KeyValuePair) -> Optional[KeyValuePair]:
    """Check if pair should be included based on config"""
    if config.exclude.keys:
        for key in pair.keypath:
            if config.exclude.keys.match(str(key)):
                logging.debug(f"filter_included excluded key '{key}'")
                return None  # Excluded key

    if config.exclude.values:
        if config.exclude.values.match(pair.value):
            logging.debug(f"filter_included excluded value '{pair.value}'")
            return None  # Excluded value

    logging.debug(f"filter_included included pair '{pair}'")
    return pair  # Included value


def filter_static(pair: KeyValuePair) -> Optional[KeyValuePair]:
    """Check if pair contains hardcoded static values"""
    pair.key = strip_string(pair.key)
    pair.value = strip_string(pair.value)

    if not is_static(pair.key, pair.value):
        logging.debug(f"filter_static excluded value '{pair.value}'")
        return None  # Dynamic value

    logging.debug(f"filter_static included value '{pair.value}'")
    return pair  # Static value


def load_plugin(file: Path) -> Optional[object]:
    """
    Loads the correct plugin for given file.
    Returns None if no plugin found.
    """
    if file.suffix in [".dist", ".template"]:
        filetype = file.stem.split(".")[-1]
    else:
        filetype = file.name.split(".")[-1]

    if filetype in ["yaml", "yml"]:
        return Yml

    elif filetype == "json":
        return Json

    elif filetype == "xml":
        return Xml

    elif filetype.startswith("npmrc"):
        return Npmrc

    elif filetype.startswith("pypirc"):
        return Pypirc

    elif file.name == "pip.conf":
        return Pip

    elif file.name == "build.gradle":
        return Gradle

    elif filetype in ["conf", "cfg", "cnf", "config", "ini", "env", "credentials", "s3cfg"]:
        return Config

    elif filetype == "properties":
        return Jproperties

    elif filetype.startswith(("sh", "bash", "zsh", "env")):
        return Shell

    elif "dockerfile" in file.name.lower():
        return Dockerfile

    elif filetype == "dockercfg":
        return Dockercfg

    elif filetype.startswith("htpasswd"):
        return Htpasswd

    elif filetype == "txt":
        return Plaintext

    elif filetype.startswith("htm"):
        return Html

    elif filetype in ["py", "py3", "py35", "py36", "py37", "py38", "py39"]:
        return Python

    elif REGEX_PRIVKEY_FILE.match(filetype):
        return Plaintext

    return None



import json
import logging
from argparse import Namespace
from itertools import chain
from os import environ
from sys import exit
from typing import Iterator

from whispers.core.args import parse_args
from whispers.core.config import load_config
from whispers.core.pairs import make_pairs
from whispers.core.rules import load_rules
from whispers.core.scope import load_scope
from whispers.core.secrets import detect_secrets
from whispers.models.pair import KeyValuePair

environ["PYTHONUTF8"] = "1"
environ["PYTHONIOENCODING"] = "UTF-8"


def main() -> None:  # pragma: no cover
    """Main entry point"""
    args = parse_args()

    secrets = []
    for secret in run(args):
        logging.warning(str(secret))
        secrets.append(secret.to_json())

    args.output.write(json.dumps(secrets) + "\n")

    exit(args.exitcode)


def run(args: Namespace) -> Iterator[KeyValuePair]:
    """Main worker process"""
    config = load_config(args)
    rules = load_rules(args, config)
    scope = load_scope(args, config)
    parsed = map(lambda file: make_pairs(config, file), scope)
    detected = map(lambda pairs: detect_secrets(rules, pairs), parsed)
    secrets = chain.from_iterable(detected)

    return secrets


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Iterator

from whispers.models.pair import KeyValuePair


class Npmrc:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        for lineno, line in enumerate(filepath.open(), 1):
            if ":_authToken=" not in line:
                continue

            value = line.split(":_authToken=")[-1].strip()
            if value:
                key = "npm authToken"
                yield KeyValuePair(key, value, line=lineno)



import re
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Pattern

from whispers.core.constants import DEFAULT_SEVERITY
from whispers.core.utils import default_rules, list_rule_prop, load_regex

DEFAULT_RULES = default_rules()


@dataclass
class Include:
    """AppConfig include configuration class"""

    files: Optional[List] = field(default_factory=lambda: ["**/*"])  # globs
    rules: Optional[List] = field(default_factory=lambda: list_rule_prop("id", DEFAULT_RULES))
    groups: Optional[List] = field(default_factory=lambda: list_rule_prop("group", DEFAULT_RULES))
    severity: Optional[List] = field(default_factory=lambda: DEFAULT_SEVERITY)


@dataclass
class Exclude:
    """AppConfig exclude configuration class"""

    files: Optional[Pattern] = field(default_factory=lambda: None)  # regex
    keys: Optional[Pattern] = field(default_factory=lambda: None)  # regex
    values: Optional[Pattern] = field(default_factory=lambda: None)  # regex
    rules: Optional[List] = field(default_factory=list)
    groups: Optional[List] = field(default_factory=list)
    severity: Optional[List] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.files = self._compile("files")
        self.keys = self._compile("keys")
        self.values = self._compile("values")

    def _compile(self, idx) -> Optional[Pattern]:
        """
        Create a single regex statement from a list,
        and compile it for efficient matching.
        """
        regex_list = self.__dict__[idx]
        if regex_list is None:
            return None

        return load_regex("|".join(regex_list), flags=re.IGNORECASE)


@dataclass
class AppConfig:
    """Main application configuration class"""

    include: Include = field(default_factory=Include)
    exclude: Exclude = field(default_factory=Exclude)

    def __init__(self, config: Dict = {}) -> None:
        self.include = Include(**config.get("include", {}))
        self.exclude = Exclude(**config.get("exclude", {}))



import json
from pathlib import Path
from typing import Iterator

from whispers.models.pair import KeyValuePair


class Dockercfg:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        config = json.loads(filepath.read_text())
        key = "auth"
        keypath = ["auths", key]

        if "auths" not in config:
            return

        for auth in config["auths"].values():
            if key not in auth:
                continue

            token = auth[key]
            yield KeyValuePair("dockercfg auth", token, keypath)



import shlex
from pathlib import Path
from typing import Iterator, List, Tuple

from whispers.core.constants import ESCAPED_CHARS
from whispers.core.utils import KeyValuePair, global_exception_handler, strip_string


class Shell:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        for cmdline, lineno in self.read_commands(filepath):
            try:
                cmd = shlex.split(cmdline)
            except Exception:
                global_exception_handler(filepath.as_posix(), cmdline)
                continue

            if not cmd:
                continue

            yield from self.variables(cmd, lineno)

            if cmd[0].lower() == "curl":
                yield from self.curl(cmd, lineno)

    def read_commands(self, filepath: Path) -> Tuple[str, int]:
        ret = []
        for lineno, line in enumerate(filepath.open(), 1):
            line = line.strip()
            if line.startswith("#"):  # Comments
                line = line.lstrip("#").strip()
                line = line.translate(ESCAPED_CHARS)

            elif line.endswith("\\"):  # Multi-line commands
                ret.append(line[:-1])
                continue

            ret.append(line)
            yield " ".join(ret), lineno
            ret = []

    def variables(self, cmd: List[str], lineno: int = 0) -> Iterator[KeyValuePair]:
        """
        Checks if Shell variables contain a hardcoded or a default value.
        Examples:
            password="defaultPassword"
            password=${ENV_VAR:-defaultPassword}
        """
        for value in cmd:
            key = ""

            if "=" in value and len(value.split("=")) == 2:
                key, value = value.split("=")  # Variable assignment

            if ":-" in value and value.startswith("${") and value.endswith("}"):
                key, value = value[2:-1].split(":-")  # Default value

            if not (key or value):
                continue

            yield KeyValuePair(key, value, line=lineno)

    def curl(self, cmd: List[str], lineno: int) -> Iterator[KeyValuePair]:
        key = "password"
        indicators_combined = ["-u", "--user", "-U", "--proxy-user", "-E", "--cert"]
        indicators_single = ["--tlspassword", "--proxy-tlspassword"]
        indicators = indicators_combined + indicators_single
        for indicator in indicators:
            if indicator not in cmd:
                continue

            idx = cmd.index(indicator)
            if len(cmd) == idx + 1:
                continue  # End of command

            credentials = strip_string(cmd[idx + 1])
            if indicator in indicators_single:
                yield KeyValuePair(key, credentials, [key])

            else:
                if ":" not in credentials:
                    continue  # Password not specified

                yield KeyValuePair(key, credentials.split(":")[1], keypath=[key], line=lineno)



import logging
from argparse import ArgumentParser, Namespace
from functools import wraps
from sys import argv, exit, stdout

from whispers.__version__ import __version__, __whispers__
from whispers.core.constants import DEFAULT_PATH
from whispers.core.utils import default_rules, load_regex


def argument_parser() -> ArgumentParser:
    """CLI argument parser"""
    args_parser = ArgumentParser("whispers", description=("Identify secrets in static structured text."))
    args_parser.add_argument("--info", action="store_true", help="show extended help and exit")
    args_parser.add_argument("--version", action="version", version=__version__, help="show version and exit")
    args_parser.add_argument("--init", default=False, action="store_true", help="make config and exit")
    args_parser.add_argument("-c", "--config", default=None, help="config filename")
    args_parser.add_argument("-o", "--output", default=None, help="output filename")
    args_parser.add_argument("-l", "--log", default=None, help="log filename")
    args_parser.add_argument(
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        default=logging.WARNING,
        help="show debug log",
    )
    args_parser.add_argument("-e", "--exitcode", default=0, type=int, help="exit code on success")
    args_parser.add_argument("-f", "--files", default=None, help="csv of globs for including files")
    args_parser.add_argument("-F", "--xfiles", default=None, help="regex for excluding files")
    args_parser.add_argument("-g", "--groups", default=None, help="csv of rule groups to report (see --info)")
    args_parser.add_argument("-G", "--xgroups", default=None, help="csv of rule groups to exclude (see --info)")
    args_parser.add_argument("-r", "--rules", default=None, help="csv of rule IDs to report (see --info)")
    args_parser.add_argument("-R", "--xrules", default=None, help="csv of rule IDs to exclude (see --info)")
    args_parser.add_argument("-s", "--severity", default=None, help="csv of severity levels to report (see --info)")
    args_parser.add_argument("-S", "--xseverity", default=None, help="csv of severity levels to exclude (see --info)")
    args_parser.add_argument("src", nargs="*", help="target file or directory")

    args_parser.print_help = show_splash(args_parser.print_help)

    return args_parser


def parse_args(arguments: list = argv[1:]) -> Namespace:
    """Parses a list into a namespace"""
    args, _ = argument_parser().parse_known_args(arguments)

    if args.info:
        show_info()
        exit()

    if args.init:
        show_config()
        exit()

    if not args.src:
        argument_parser().print_help()
        exit()
    else:
        args.src = args.src[0]

    if args.log:
        args.log = logging.FileHandler(args.log, mode="w")
    else:
        args.log = logging.StreamHandler()

    logging.basicConfig(format="[%(levelname)s] %(message)s", level=args.debug, handlers=[args.log])

    if args.output:
        args.output = open(args.output, "w")
    else:
        args.output = stdout

    if args.files:
        args.files = args.files.split(",")

    if args.xfiles:
        args.xfiles = load_regex(args.xfiles)

    if args.rules:
        args.rules = args.rules.split(",")

    if args.xrules:
        args.xrules = args.xrules.split(",")

    if args.groups:
        args.groups = args.groups.split(",")

    if args.xgroups:
        args.xgroups = args.xgroups.split(",")

    if args.severity:
        args.severity = args.severity.split(",")

    if args.xseverity:
        args.xseverity = args.xseverity.split(",")

    return args


def show_splash(func, **kwargs):
    @wraps(func)
    def splash(*args, **kwargs):
        print(__whispers__)
        print(__version__.rjust(64), end="\n\n")
        return func(*args, **kwargs)

    return splash


def show_config():
    print(DEFAULT_PATH.joinpath("config.yml").read_text())


def show_info():
    argument_parser().print_help()
    rules_table = []
    col_width = 20
    for rule in default_rules():
        line = (
            "    "
            + rule["group"].ljust(col_width)[:col_width]
            + " | "
            + rule["id"].ljust(col_width)[:col_width]
            + " | "
            + rule["severity"].ljust(col_width)[:col_width]
        )
        rules_table.append(line)

    draw_line = "\n  " + ("+--" + "-" * col_width) * 3 + "+"
    print(
        "\n\nrules:\n"
        + draw_line
        + "\n  | "
        + "group".ljust(col_width)
        + " | "
        + "rule id".ljust(col_width)
        + " | "
        + "severity".ljust(col_width)
        + " |"
        + draw_line
        + "\n"
        + "\n".join(sorted(rules_table))
        + draw_line
        + "\n\nreadme:  https://github.com/adeptex/whispers\n"
    )



import re
from pathlib import Path

DEFAULT_PATH = Path(__file__).parents[1]

DEFAULT_SEVERITY = ["Critical", "High", "Medium", "Low", "Info"]

ESCAPED_CHARS = str.maketrans({"'": r"\'", '"': r"\""})

REGEX_URI = re.compile(r"[:\w\d]+://.+", flags=re.IGNORECASE)
REGEX_PATH = re.compile(r"^((([A-Z]|file|root):)?(\.+)?[/\\]+).*$", flags=re.IGNORECASE)
REGEX_IAC = re.compile(r"\![A-Za-z]+ .+", flags=re.IGNORECASE)
REGEX_PRIVKEY_FILE = re.compile(r"(rsa|dsa|ed25519|ecdsa|pem|crt|cer|ca-bundle|p7b|p7c|p7s|ppk|pkcs12|pfx|p12)")
REGEX_ENVVAR = re.compile(r"^\$\$?\{?[A-Z0-9_]+\}?$")



from importlib import import_module
from setuptools import find_packages, setup
from pathlib import Path
from sys import version_info


install_requires = [
    "astroid",
    "beautifulsoup4",
    "crossplane",
    "jellyfish",
    "jproperties",
    "luhn",
    "lxml",
    "pyyaml",
]

dev_requires = [
    "autoflake~=1.4",
    "autopep8~=1.7",
    "black==22.8; python_version <= '3.6'",
    "black~=24.3; python_version > '3.6'",
    "build~=0.8",
    "coverage~=4.5",
    "coverage-badge~=1.0",
    "flake8~=5.0",
    "isort~=5.9",
    "pytest~=7.0",
    "pytest-mock~=3.6",
    "pip-tools~=6.2",
    "wheel~=0.37",
    "twine~=3.4",
]


def get_version():
    return import_module("whispers.__version__").__version__


def get_readme():
    return Path(__file__).parent.joinpath("README.md").read_text()


setup(
    name="whispers",
    version=get_version(),
    url="https://github.com/adeptex/whispers",
    author="Artëm Tsvetkov",
    author_email="adeptex@users.noreply.github.com",
    description="Identify secrets in static structured text",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests", "tests.*")),
    include_package_data=True,
    platforms="any",
    install_requires=install_requires,
    setup_requires=["pytest-runner"],
    tests_require=dev_requires,
    extras_require={"dev": dev_requires},
    entry_points={"console_scripts": ["whispers=whispers.main:main"]},
    classifiers=[
        "License :: OSI Approved :: BSD License",
    ],
)
